﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Microsoft.Office.Interop.Excel;
using DocumentFormat.OpenXml;
using DocumentFormat.OpenXml.Packaging;
using DocumentFormat.OpenXml.Spreadsheet;
using System.IO;
using MasterConsole.Code;
using System.Diagnostics;
using edu.stanford.nlp.ie.crf;
using java.util;
using edu.stanford.nlp.ling;
using edu.stanford.nlp.process;


namespace MasterConsole
{
    class Program
    {
        public static Stopwatch HammingWatch;
        public static Stopwatch JaccardWatch;
        public static Stopwatch JaroWatch;
        public static Stopwatch JaroWinklerWatch;
        public static Stopwatch LevenshteinWatch;
        public static Stopwatch LCSubsequenceWatch;
        public static Stopwatch LCSubstringWatch;
        public static Stopwatch NLevenshteinWatch;
        public static Stopwatch ROWatch;
        public static Stopwatch SDWatch;
        public static Stopwatch TanimotoWatch;

        //public enum FuzzyAlgorithm
        //{
        //    HammingDistance,
        //    JaccardDistance,
        //    JaroDistance,
        //    JaroWinklerDistance,
        //    LevenshteinDistance,
        //    LongestCommonSubsequence,
        //    LongestCommonSubstring,
        //    OverlapCoefficient,
        //    RatcliffObershelpSimilarity,
        //    SorensenDiceDistance,
        //    TanimotoCoefficient
        //}

        static void Main(string[] args)
        {
            var inputExcelFilePath = "C:\\Dropbox\\Thomas\\GDF SUEZ Org Chart Details with Errors_25012015.xlsx";
            //LevenshteinDistanceTest(); // Uses Levenshtein Distance and gazetteer lists to classify words that might have spelling errors.
            NERTest(inputExcelFilePath); // This is the best one. It doesn't seem to care about spelling errors that much and therefor does not need fuzzy algorithms.
            //NERTest(@"Manager North Sea/Haltenbanken
//");
              
        }

        /// <summary>
        /// http://stackoverflow.com/questions/4079135/splitting-a-string-with-uppercase
        /// </summary>
        /// <param name="input"></param>
        /// <returns></returns>
        public static string SplitOnCase(string input)
        {
            string output = "";

            foreach (char letter in input)
            {
                if (Char.IsUpper(letter) && output.Length > 0)
                    output += " " + letter;
                else
                    output += letter;
            }
            return output;
        }

        /// <summary>
        /// http://sharpnlp.codeplex.com/
        /// </summary>
        public static void SharpNLPTest() 
        {

        }
        /// <summary>
        /// This is a test-method and is not very generic yet
        /// </summary>
        /// <param name="filepath">The path of the excel file that contains strings to be classified</param>
        /// <param name="testString">If provided, it classifies the string as a person, organisation or location</param>
        public static void NERTest(string filepath, string testString = "")
        {
            CRFClassifier Classifier = CRFClassifier.getClassifierNoExceptions(@"C:\Workspace Visual Studio\english.all.3class.distsim.crf.ser.gz"); // This list can be downloaded from the Stanfords NER website. There are also 7class files which classifies for Time, Location, Organization, Person, Money, Percent and Date (need to look into this).
            if (testString != String.Empty) {
                //Console.WriteLine(testString.Trim().Trim(new Char[] { '(', ')' }));
                Console.WriteLine("{0}\n", Classifier.classifyToString(testString.Replace("/", " ").Replace("(" , "").Replace(")", "")));
                //edu.stanford.nlp.process.PTBTokenizer.PTBTokenizerFactory test = new edu.stanford.nlp.process.PTBTokenizer.PTBTokenizerFactory();
                //PTBTokenizer ptbt = new PTBTokenizer(, new CoreLabelTokenFactory(), "");
                Console.ReadLine();
                return;
            }
            var temp = "";

            #region Test data
            var firstColumn = GetColumnList(1, 1, filepath);
            var secondColumn = GetColumnList(1, 2, filepath);
            var thirdColumn = GetColumnList(1, 3, filepath);
            var allColumns = firstColumn.Concat(secondColumn).Concat(thirdColumn);
            var tests = allColumns.Take(5);
            //startTime = DateTime.Now;
            //var lastnames2 = ReadExcelFileSAX(filename);
            #endregion
            var classifStr = "";
            var fuzzyResList = new List<FuzzyResult>();
            var fuzzyRes = new FuzzyResult();
            var personTagCount = 0;
            var orgTagCount = 0;
            var trimmedStr = "";
            foreach (var test in allColumns)
            {
                trimmedStr = test.Trim().Replace("(", "").Replace(")", "");
                personTagCount = 0;
                orgTagCount = 0;
                fuzzyRes = new FuzzyResult();
                fuzzyRes.OriginalString = test;
                //Console.WriteLine("{0}\n", Classifier.classifyToString(test));
                foreach (List sentence in Classifier.classify(trimmedStr).toArray()) //Main loop for classification
                {
                    foreach (CoreLabel word in sentence.toArray()) //Each name/position title is seperated into words
                    {
                        classifStr = (string)word.get(new CoreAnnotations.AnswerAnnotation().getClass()); //Classifies the word
                        Console.WriteLine("{0}/{1} ", word.word(), classifStr);
                        //Console.WriteLine();
                        if (classifStr == "PERSON")
                            personTagCount++;
                        else if (classifStr == "O")
                            orgTagCount++;
                    }
                }
                if (personTagCount > orgTagCount) // If most of the words are classified as person names it classifies the "sentance" (name or position title)
                    fuzzyRes.Classification = Classification.Name;
                else if (orgTagCount > 0) // Else it's classified as a position title
                    fuzzyRes.Classification = Classification.PositionTitle;


                if (fuzzyRes.Classification == Classification.PositionTitle || fuzzyRes.Classification == Classification.Empty) // An extra check for space errors. For example "AnnaSvendsen" will be split into "Anna Svendsen" and then evaluated.
                {
                    personTagCount = 0;
                    orgTagCount = 0;
                    temp = SplitOnCase(trimmedStr);
                    foreach (List sentence in Classifier.classify(temp).toArray())
                    {
                        foreach (CoreLabel word in sentence.toArray())
                        {
                            classifStr = (string)word.get(new CoreAnnotations.AnswerAnnotation().getClass());
                            Console.WriteLine("{0}/{1} ", word.word(), word.get(new CoreAnnotations.AnswerAnnotation().getClass()));
                            //Console.WriteLine();
                            if (classifStr == "PERSON")
                                personTagCount++;
                            else if (classifStr == "O")
                                orgTagCount++;
                        }
                    }
                    if (personTagCount > orgTagCount)
                        fuzzyRes.Classification = Classification.Name;
                    else if (orgTagCount > 0)
                        fuzzyRes.Classification = Classification.PositionTitle;
                }
                fuzzyResList.Add(fuzzyRes);

            }


            //Console.WriteLine("{0}\n", Classifier.classifyToString("AnneSvendsen"));

            Console.WriteLine("Writing to file...");
            using (var file = File.CreateText("c:\\NERtest4.csv")) // Writes the results to file
            {
                file.WriteLine("OriginalString;Classification");
                //file.WriteLine("OriginalString;HammingDistanceBestMatch;JaccardDistanceBestMatch;JaroDistanceBestMatch;JaroWinklerDistanceBestMatch;LevenshteinDistanceBestMatch;LongestCommonSubsequenceBestMatch;LongestCommonSubstringBestMatch;OverlapCoefficientBestMatch;RatcliffObershelpSimilarityBestMatch;SorensenDiceDistanceBestMatch;TanimotoCoefficientBestMatch");
                foreach (var result in fuzzyResList)
                {
                    file.WriteLine(result.OriginalString + ";" + result.Classification);
                }
            }
            Console.WriteLine("Done.");
            //const string S1 = "Good afternoon Rajat Raina, how are you today?";
            //const string S2 = "I go to school at Stanford University, which is located in California.";
            //Console.WriteLine("{0}\n", Classifier.classifyToString(S1));
            //Console.WriteLine("{0}\n", Classifier.classifyWithInlineXML(S2));
            //Console.WriteLine("{0}\n", Classifier.classifyToString(S2, "xml", true));
            Console.ReadLine();
        }

        public static void LevenshteinDistanceTest()
        {
            Stopwatch stopwatch = new Stopwatch();
            stopwatch.Start();
            HammingWatch = new Stopwatch();
            JaccardWatch = new Stopwatch();
            JaroWatch = new Stopwatch();
            JaroWinklerWatch = new Stopwatch();
            LevenshteinWatch = new Stopwatch();
            LCSubsequenceWatch = new Stopwatch();
            LCSubstringWatch = new Stopwatch();
            //NLevenshteinWatch = new Stopwatch();
            ROWatch = new Stopwatch();
            SDWatch = new Stopwatch();
            TanimotoWatch = new Stopwatch();
            var filename = "C:\\Dropbox\\Thomas\\Names and Resource Types Gazeteer Lists_22012015.xlsx";

            #region Firstname-lists
            var norFirstNames = GetColumnList(1, 1, filename);
            var firtNames = GetColumnList(1, 3, filename);
            //var emailFirstnames = GetColumnList(1, 5, filename);
            var dutchFirstNames = GetColumnList(1, 7, filename);
            var allFirstNames = new GazetteerList(Classification.FirstName, norFirstNames.Concat(firtNames).Concat(dutchFirstNames).Distinct());
            #endregion

            #region Lastname-lists
            var norLastNames = GetColumnList(2, 1, filename);
            var lastNames = GetColumnList(2, 3, filename);
            //var emailLastnames = GetColumnList(2, 5, filename);
            var allLastNames = new GazetteerList(Classification.LastName, norLastNames.Concat(lastNames).Distinct());
            #endregion

            #region Position titles
            var oilCompPosTitles = new GazetteerList(Classification.PositionTitle, GetColumnList(3, 1, filename));
            //var posTitlesLarge = GetColumnList(3, 3, filename);
            //var categories = GetColumnList(3, 5, filename);
            //var resourceTypes = GetColumnList(3, 6, filename);
            #endregion

            filename = "C:\\Dropbox\\Thomas\\GDF SUEZ Org Chart Details with Errors_25012015.xlsx";

            #region Test data
            var firstColumn = GetColumnList(1, 1, filename);
            var secondColumn = GetColumnList(1, 2, filename);
            var thirdColumn = GetColumnList(1, 3, filename);
            var allColumns = firstColumn.Concat(secondColumn).Concat(thirdColumn);
            //startTime = DateTime.Now;
            //var lastnames2 = ReadExcelFileSAX(filename);
            #endregion

            #region Fuzzy String Matching Library (not that good)
            //var split = "";
            //StringMatching.FuzzyStringMatching test = new StringMatching.FuzzyStringMatching();
            //var array = allFirstnames.ToArray();
            //foreach (var row in secondColumn) {
            //    Console.WriteLine("Evaluating '" + row + "'...");
            //    split = row;
            //    if (row.Contains(' '))
            //        split = row.Split(' ').First();
            //    if (split == string.Empty)
            //        continue;
            //    var resultSet = test.GetSimiarities(split, array);
            //    foreach (var result in resultSet.Where(rs => rs.Result >= 0.8))
            //    {
            //        Console.WriteLine("Found word for '" + split + "': '" + result.Word + "' with result: " + result.Result);
            //    }
            //}
            #endregion

            Console.WriteLine("Starting to analyze...");
            var completeList = FuzzyMatch(allColumns, oilCompPosTitles, allFirstNames, allLastNames);
            Console.WriteLine("Complete.");
            //Console.WriteLine();
            //foreach (var bestMatch in bestMatches)
            //{
            //    Console.WriteLine("originalString: " + bestMatch.OriginalString + ", bestMatch: " + bestMatch.BestMatch + ", classification: " + bestMatch.Classification);
            //}

            #region write to csv
            Console.WriteLine("Writing to file...");
            using (var file = File.CreateText("c:\\test5.csv"))
            {
                file.WriteLine("OriginalString;LevenshteinDistanceBestMatch;Classification;Value");
                //file.WriteLine("OriginalString;HammingDistanceBestMatch;JaccardDistanceBestMatch;JaroDistanceBestMatch;JaroWinklerDistanceBestMatch;LevenshteinDistanceBestMatch;LongestCommonSubsequenceBestMatch;LongestCommonSubstringBestMatch;OverlapCoefficientBestMatch;RatcliffObershelpSimilarityBestMatch;SorensenDiceDistanceBestMatch;TanimotoCoefficientBestMatch");
                foreach (var result in completeList)
                {
                    file.WriteLine(result.OriginalString + ";" + result.BestMatch + ";" + result.Classification + ";" + result.Value);
                }
            }
            #endregion
            Console.WriteLine("Done.");


            //IEnumerable<T> sequence;
            //string csv = String.Join(",", sequence);

            //Console.WriteLine("First three objects of the column '" + list.GetValue(0) + "':");
            //for (int i = 1; i < 4; i++)
            //{
            //    Console.WriteLine(list.GetValue(i));
            //}
            //Console.WriteLine("Count of the column '" + norFirstnames.First() + "':" + norFirstnames.Count());
            //Console.WriteLine("Count of the column '" + firtnames.First() + "':" + firtnames.Count());
            //Console.WriteLine("Count of the column '" + emailFirstnames.First() + "':" + emailFirstnames.Count());
            //Console.WriteLine("Count of the column '" + dutchFirstnames.First() + "':" + dutchFirstnames.Count());

            //Console.WriteLine("Count of the column '" + norLastnames.First() + "':" + norLastnames.Count());
            //Console.WriteLine("Count of the column '" + lastnames.First() + "':" + lastnames.Count());
            //Console.WriteLine("Count of the column '" + emailLastnames.First() + "':" + emailLastnames.Count());

            //Console.WriteLine("Count of the column '" + oilCompPosTitles.First() + "':" + oilCompPosTitles.Count());
            //Console.WriteLine("Count of the column '" + posTitlesLarge.First() + "':" + posTitlesLarge.Count());
            //Console.WriteLine("Count of the column '" + categories.First() + "':" + categories.Count());
            //Console.WriteLine("Count of the column '" + resourceTypes.First() + "':" + resourceTypes.Count());

            //Console.WriteLine("Count of the column '" + firstColumn.First() + "':" + firstColumn.Count());
            //Console.WriteLine("Count of the column '" + secondColumn.First() + "':" + secondColumn.Count());
            //Console.WriteLine("Matches found: " + found + " of " + secondColumnCount + " (" + ((double)found / (double)secondColumnCount) * 100 + "%)");
            //Console.WriteLine("Approx matches found: " + approxFound + " of " + secondColumnCount + " (" + ((double)approxFound / (double)secondColumnCount) * 100 + "%)");
            //Console.WriteLine("Total matches found: " + bestMatches.Where(b => b.BestMatch != string.Empty).Count() + " of " + secondColumnCount + " (" + ((double)(bestMatches.Where(b => b.BestMatch != string.Empty).Count()) / (double)secondColumnCount) * 100 + "%)");
            //Console.WriteLine("Time used: " + (DateTime.Now - startTime).Milliseconds + " ms");
            stopwatch.Stop();

            //Console.WriteLine("Hamming Distance time used: " + (HammingWatch.ElapsedMilliseconds / 1000) + " seconds");
            //Console.WriteLine("Jaccard Distance time used: " + (JaccardWatch.ElapsedMilliseconds / 1000) + " seconds");
            //Console.WriteLine("Jaro Distance time used: " + (JaroWatch.ElapsedMilliseconds / 1000) + " seconds");
            //Console.WriteLine("Jaro-Winkler Distance time used: " + (JaroWinklerWatch.ElapsedMilliseconds / 1000) + " seconds");
            Console.WriteLine("Levenshtein Distance time used: " + (LevenshteinWatch.ElapsedMilliseconds / 1000) + " seconds");
            //Console.WriteLine("Longest Common Subsequence time used: " + (LCSubsequenceWatch.ElapsedMilliseconds / 1000) + " seconds");
            //Console.WriteLine("Longest Common Substring time used: " + (LCSubstringWatch.ElapsedMilliseconds / 1000) + " seconds");
            //Console.WriteLine("Normalized Levenshtein Distance time used: " + (NLevenshteinWatch.ElapsedMilliseconds / 1000) + " seconds");
            //Console.WriteLine("Ratcliff-Obershelp Similarity time used: " + (ROWatch.ElapsedMilliseconds / 1000) + " seconds");
            //Console.WriteLine("Sørensen-Dice Distance time used: " + (SDWatch.ElapsedMilliseconds / 1000) + " seconds");
            //Console.WriteLine("Tanimoto Coefficient time used: " + (TanimotoWatch.ElapsedMilliseconds / 1000) + " seconds");

            Console.WriteLine("Total time used: " + (stopwatch.ElapsedMilliseconds / 1000) + " seconds");
            Console.WriteLine("Stop.");
            Console.ReadLine();
        }

        public static double NormalizedHammingDistance(string source, string target)
        {
            int maxL = Math.Max(source.Length, target.Length);
            int dist = FuzzyString.ComparisonMetrics.HammingDistance(source, target);
            return 1.0 - (double)dist / (double)maxL;
        }

        /// <summary>
        /// Takes a list of strings to be fuzzy matched against gazetteer lists with certain classifiers. 
        /// The strings are then classified based on the best match of the gazetteer lists.
        /// </summary>
        /// <param name="testingList"></param>
        /// <param name="gazetteerLists"></param>
        /// <returns>A list of fuzzy results</returns>
        public static List<FuzzyResult> FuzzyMatch(IEnumerable<string> testingList, params GazetteerList[] gazetteerLists)
        {
            var temp = "";
            var tempValue = -1.0;
            var returningList = new List<FuzzyResult>();
            //var fuzzyRes = new CompleteFuzzyResult();
            var fuzzyRes = new FuzzyResult();
            var fuzzyRes2 = new FuzzyResult();
            var testingListCount = testingList.Count();
            var firstName = "";
            var lastName = "";
            var trimmedElem = "";
            var i = 0;
            IEnumerable<string> split;
            var firstNameList = gazetteerLists.Where(g => g.Classification == Classification.FirstName).FirstOrDefault();
            var lastNameList = gazetteerLists.Where(g => g.Classification == Classification.LastName).FirstOrDefault();
            var hasFirstNameList = firstNameList != null;
            var hasLastNameList = lastNameList != null;
            var searchForName = hasFirstNameList && hasLastNameList;
            //var comparer = new HammingDistanceComparer();
            //comparer.MaxDist = 2;
            foreach (var elem in testingList)
            {
                trimmedElem = elem.Trim();
                Console.Title = (int)(((double)i / (double)testingListCount) * 100) + "% - MasterConsole";
                Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                Console.WriteLine("Checking for: " + trimmedElem);
                //fuzzyRes = new CompleteFuzzyResult();
                fuzzyRes = new FuzzyResult();
                fuzzyRes.OriginalString = trimmedElem;


                #region Comparing for name
                if (searchForName)
                {
                    FuzzyNameResult fnr = new FuzzyNameResult();
                    split = trimmedElem.Split(' ').Where(s => s != "");
                    if (split.Count() > 1) // TODO: Should also check for space errors here, could have a middlename but they f.ex. forgot space between middlename and firstname
                    {
                        #region First name comparison (should be done in an own method that takes in list, string source, string target and outputs FuzzyResult)
                        fuzzyRes2 = new FuzzyResult();
                        firstName = split.First();
                        //fuzzyRes2.BestMatch = firstNameList.List.OrderByDescending(l => NormalizedHammingDistance(firstName, l)).FirstOrDefault() ?? "";

                        LevenshteinWatch.Start();
                        fuzzyRes2.BestMatch = firstNameList.List.OrderByDescending(l => LevenshteinDistance.ComputeNormalized(firstName, l)).FirstOrDefault() ?? "";
                        LevenshteinWatch.Stop();
                        fuzzyRes2.Classification = Classification.FirstName;
                        fuzzyRes2.OriginalString = firstName;
                        fuzzyRes2.Value = LevenshteinDistance.ComputeNormalized(firstName, fuzzyRes2.BestMatch);
                        fnr.FirstName = fuzzyRes2;
                        #endregion
                        #region Last name comparison
                        fuzzyRes2 = new FuzzyResult();
                        lastName = split.Last();
                        LevenshteinWatch.Start();
                        fuzzyRes2.BestMatch = lastNameList.List.OrderByDescending(l => LevenshteinDistance.ComputeNormalized(lastName, l)).FirstOrDefault() ?? "";
                        LevenshteinWatch.Stop();
                        fuzzyRes2.Classification = Classification.LastName;
                        fuzzyRes2.OriginalString = lastName;
                        fuzzyRes2.Value = LevenshteinDistance.ComputeNormalized(lastName, fuzzyRes2.BestMatch);
                        fnr.LastName = fuzzyRes2;
                        #endregion
                        #region Middle names comparison (use both middle name and last name lists as the middle name could be a last name)
                        foreach (var middlename in split.Skip(1).Take(split.Count() - 2))
                        {
                            fuzzyRes2 = new FuzzyResult();
                            LevenshteinWatch.Start();
                            fuzzyRes2.BestMatch = lastNameList.List.OrderByDescending(l => LevenshteinDistance.ComputeNormalized(middlename, l)).FirstOrDefault() ?? ""; // TODO: Include middle name list when it exists
                            LevenshteinWatch.Stop();
                            fuzzyRes2.Classification = Classification.LastName;
                            fuzzyRes2.OriginalString = middlename;
                            fuzzyRes2.Value = LevenshteinDistance.ComputeNormalized(middlename, fuzzyRes2.BestMatch);
                            fnr.MiddleNames.Add(fuzzyRes2);
                        }
                        #endregion
                        fuzzyRes.BestMatch = fnr.GetBestNameMatch;
                        fuzzyRes.Classification = Classification.Name;
                        fuzzyRes.Value = fnr.GetAverageValue;

                    }
                    else
                    {
                        //Code for handling space errors here.
                    }
                }
                #endregion

                foreach (var list in gazetteerLists)
                {
                    if (fuzzyRes.Value == 1)
                        break;
                    //if (list.List.Contains(elem))
                    //{
                    //    fuzzyRes.ContainsMatch = elem;
                    //    returningList.Add(fuzzyRes);
                    //    break;
                    //}
                    #region Hamming Distance (0 to unlim)
                    Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    Console.WriteLine("Running Levenshtein Distance...");
                    LevenshteinWatch.Start();
                    temp = list.List.OrderByDescending(l => LevenshteinDistance.ComputeNormalized(trimmedElem, l)).FirstOrDefault() ?? ""; //Will be faster to iterate with for-loop and stop if distance is 1, this is just a shorter code for now.
                    tempValue = LevenshteinDistance.ComputeNormalized(trimmedElem, temp);
                    if (tempValue > fuzzyRes.Value)
                    {
                        //fuzzyRes.HammingDistanceBestMatch = temp;
                        fuzzyRes.BestMatch = temp;
                        fuzzyRes.Classification = list.Classification;
                        fuzzyRes.Value = tempValue;
                    }
                    LevenshteinWatch.Stop();
                    #endregion
                }

                returningList.Add(fuzzyRes);

                i++;
                #region Last names and first names, needs to be split before fuzzy searching
                //foreach (var list in gazetteerLists.Where(g => g.Classification == "Last name" || g.Classification == "First name"))
                //{
                //    temp = elem;
                //    if (list.Classification == "Last name" && elem.Contains(' ') && elem.Split(' ').Last() != "")
                //        temp = elem.Split(' ').Last();
                //    if (list.Classification == "First name" && elem.Contains(' ') && elem.Split(' ').First() != "")
                //        temp = elem.Split(' ').First();
                //    foreach (var testElem in list.List)
                //    {
                //        dist = FuzzyString.ComparisonMetrics.HammingDistance(temp, testElem);
                //        if (dist == 0)
                //        {
                //            fuzzyRes.OriginalString = temp;
                //            fuzzyRes.BestMatch = testElem;
                //            fuzzyRes.Classification = list.Classification;
                //            returningList.Add(fuzzyRes);
                //            found = true;
                //            break;
                //        }
                //        if (dist <= 2)
                //        {
                //        }


                //    }
                //}
                #endregion
            }

            return returningList;
            
        }

        /// <summary>
        /// Takes a list of strings to be fuzzy matched against gazetteer lists with certain classifiers. 
        /// The strings are then classified based on the best match of the gazetteer lists.
        /// </summary>
        /// <param name="testingList"></param>
        /// <param name="gazetteerLists"></param>
        /// <returns>A list of fuzzy results</returns>
        public static List<CompleteFuzzyResult> FuzzyMatchTesting(IEnumerable<string> testingList, params GazetteerList[] gazetteerLists)
        {
            bool found;
            var temp = "";
            var temp2 = "";
            var bestMatch = "";
            var dist = 0;
            var returningList = new List<CompleteFuzzyResult>();
            var fuzzyRes = new CompleteFuzzyResult();
            //var fuzzyRes = new FuzzyResult();
            var testingListCount = testingList.Count();
            var i = 0;
            //var comparer = new HammingDistanceComparer();
            //comparer.MaxDist = 2;
            foreach (var elem in testingList)
            {
                Console.Title = (int)(((double)i / (double)testingListCount) * 100) + "% - MasterConsole";
                Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                Console.WriteLine("Checking for: " + elem);
                fuzzyRes = new CompleteFuzzyResult();
                //fuzzyRes = new FuzzyResult();
                fuzzyRes.OriginalString = elem;

                //found = false;
                foreach (var list in gazetteerLists)
                {
                    //if (list.List.Contains(elem))
                    //{
                    //    fuzzyRes.ContainsMatch = elem;
                    //    returningList.Add(fuzzyRes);
                    //    break;
                    //}
                    #region Hamming Distance (0 to unlim)
                    Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    Console.WriteLine("Running Hamming Distance...");
                    HammingWatch.Start();
                    temp = list.List.OrderBy(l => FuzzyString.ComparisonMetrics.HammingDistance(elem, l)).FirstOrDefault() ?? "";
                    if (FuzzyString.ComparisonMetrics.HammingDistance(elem, temp) < FuzzyString.ComparisonMetrics.HammingDistance(elem, fuzzyRes.HammingDistanceBestMatch))
                    {
                        fuzzyRes.HammingDistanceBestMatch = temp;
                        //fuzzyRes.BestMatch = temp;
                        //fuzzyRes.Classification = list.Classification;
                    }
                    HammingWatch.Stop();
                    #endregion

                    #region Jaccard Distance (0 to 1) (got no results, maybe not working with the FuzzyString-method)
                    //Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    //Console.WriteLine("Running Jaccard Distance...");
                    //JaccardWatch.Start();
                    //temp = list.List.OrderByDescending(l => FuzzyString.ComparisonMetrics.JaccardDistance(elem, l)).FirstOrDefault() ?? "";
                    //if (FuzzyString.ComparisonMetrics.JaccardDistance(elem, temp) > FuzzyString.ComparisonMetrics.JaccardDistance(elem, fuzzyRes.JaccardDistanceBestMatch))
                    //{
                    //    fuzzyRes.JaccardDistanceBestMatch = temp;
                    //}
                    //JaroWatch.Stop();
                    #endregion

                    #region Jaro Distance (0 to 1)
                    //Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    //Console.WriteLine("Running Jaro Distance...");
                    //JaroWatch.Start();
                    //temp = list.List.OrderByDescending(l => FuzzyString.ComparisonMetrics.JaroDistance(elem, l)).FirstOrDefault() ?? "";
                    //if (FuzzyString.ComparisonMetrics.JaroDistance(elem, temp) > FuzzyString.ComparisonMetrics.JaroDistance(elem, fuzzyRes.JaroDistanceBestMatch))
                    //{
                    //    fuzzyRes.JaroDistanceBestMatch = temp;
                    //}
                    //JaroWatch.Stop();
                    #endregion

                    #region Jaro-Winkler Distance (0 to 1)
                    //Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    //Console.WriteLine("Running Jaro-Winkler Distance...");
                    //JaroWinklerWatch.Start();
                    //temp = list.List.OrderByDescending(l => FuzzyString.ComparisonMetrics.JaroWinklerDistance(elem, l)).FirstOrDefault() ?? "";
                    //if (FuzzyString.ComparisonMetrics.JaroWinklerDistance(elem, temp) > FuzzyString.ComparisonMetrics.JaroWinklerDistance(elem, fuzzyRes.JaroWinklerDistanceBestMatch))
                    //{
                    //    fuzzyRes.JaroWinklerDistanceBestMatch = temp;
                    //}
                    //JaroWinklerWatch.Stop();
                    #endregion

                    #region Levenshtein Distance (0 to unlim)
                    //Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    //Console.WriteLine("Running Levenshtein Distance...");
                    //LevenshteinWatch.Start();
                    //temp = list.List.OrderBy(l => LevenshteinDistance.Compute(elem, l)).FirstOrDefault() ?? "";
                    //if (LevenshteinDistance.Compute(elem, temp) < LevenshteinDistance.Compute(elem, fuzzyRes.LevenshteinDistanceBestMatch))
                    //{
                    //    fuzzyRes.LevenshteinDistanceBestMatch = temp;
                    //}
                    //LevenshteinWatch.Stop();
                    #endregion

                    #region Longest Common Subsequence (string)
                    //Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    //Console.WriteLine("Running Longest Common Subsequence...");
                    //LCSubsequenceWatch.Start();
                    //temp = list.List.OrderByDescending(l => FuzzyString.ComparisonMetrics.LongestCommonSubsequence(elem, l).Length).FirstOrDefault() ?? "";
                    //if (FuzzyString.ComparisonMetrics.LongestCommonSubsequence(elem, temp).Length > FuzzyString.ComparisonMetrics.LongestCommonSubsequence(elem, fuzzyRes.LongestCommonSubsequenceBestMatch).Length)
                    //{
                    //    fuzzyRes.LongestCommonSubsequenceBestMatch = temp;
                    //}
                    //LCSubsequenceWatch.Stop();
                    #endregion

                    #region Longest Common Substring
                    //Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    //Console.WriteLine("Running Longest Common Substring...");
                    //LCSubstringWatch.Start();
                    //temp = list.List.OrderByDescending(l => FuzzyString.ComparisonMetrics.LongestCommonSubstring(elem, l).Length).FirstOrDefault() ?? "";
                    ////temp2 = FuzzyString.ComparisonMetrics.LongestCommonSubstring(elem, fuzzyRes.LongestCommonSubstringBestMatch);
                    //if (fuzzyRes.LongestCommonSubstringBestMatch == string.Empty || FuzzyString.ComparisonMetrics.LongestCommonSubstring(elem, temp).Length > FuzzyString.ComparisonMetrics.LongestCommonSubstring(elem, fuzzyRes.LongestCommonSubstringBestMatch).Length)
                    //{
                    //    fuzzyRes.LongestCommonSubstringBestMatch = temp;
                    //}
                    //LCSubstringWatch.Stop();
                    #endregion

                    #region Normalized Levenshtein Distance (did not seem to be working)
                    //Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    //Console.WriteLine("Running Normalized Levenshtein Distance...");
                    //NLevenshteinWatch.Start();
                    //temp = list.List.OrderByDescending(l => FuzzyString.ComparisonMetrics.NormalizedLevenshteinDistance(elem, l)).FirstOrDefault() ?? "";
                    //if (FuzzyString.ComparisonMetrics.NormalizedLevenshteinDistance(elem, temp) > FuzzyString.ComparisonMetrics.NormalizedLevenshteinDistance(elem, fuzzyRes.NormalizedLevenshteinDistanceBestMatch))
                    //{
                    //    fuzzyRes.NormalizedLevenshteinDistanceBestMatch = temp;
                    //}
                    //NLevenshteinWatch.Stop();
                    #endregion

                    #region Ratcliff-Obershelp Similarity
                    //Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    //Console.WriteLine("Running Ratcliff-Obershelp Similarity...");
                    //ROWatch.Start();
                    //temp = list.List.OrderByDescending(l => FuzzyString.ComparisonMetrics.RatcliffObershelpSimilarity(elem, l)).FirstOrDefault() ?? "";
                    //if (FuzzyString.ComparisonMetrics.RatcliffObershelpSimilarity(elem, temp) > FuzzyString.ComparisonMetrics.RatcliffObershelpSimilarity(elem, fuzzyRes.RatcliffObershelpSimilarityBestMatch))
                    //{
                    //    fuzzyRes.RatcliffObershelpSimilarityBestMatch = temp;
                    //}
                    //ROWatch.Stop();
                    #endregion

                    #region Sørensen-Dice Distance (got no results, maybe not working with the FuzzyString-method)
                    //Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    //Console.WriteLine("Running Sørensen-Dice Distance...");
                    //SDWatch.Start();
                    //temp = list.List.OrderByDescending(l => FuzzyString.ComparisonMetrics.SorensenDiceDistance(elem, l)).FirstOrDefault() ?? "";
                    //if (FuzzyString.ComparisonMetrics.SorensenDiceDistance(elem, temp) > FuzzyString.ComparisonMetrics.SorensenDiceDistance(elem, fuzzyRes.SorensenDiceDistanceBestMatch))
                    //{
                    //    fuzzyRes.SorensenDiceDistanceBestMatch = temp;
                    //}
                    //SDWatch.Stop();
                    #endregion

                    #region Tanimoto Coefficient
                    //Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                    //Console.WriteLine("Running Tanimoto Coefficient...");
                    //TanimotoWatch.Start();
                    //temp = list.List.OrderByDescending(l => FuzzyString.ComparisonMetrics.TanimotoCoefficient(elem, l)).FirstOrDefault() ?? "";
                    //if (FuzzyString.ComparisonMetrics.TanimotoCoefficient(elem, temp) > FuzzyString.ComparisonMetrics.TanimotoCoefficient(elem, fuzzyRes.TanimotoCoefficientBestMatch))
                    //{
                    //    fuzzyRes.TanimotoCoefficientBestMatch = temp;
                    //}
                    //TanimotoWatch.Stop();
                    #endregion
                    returningList.Add(fuzzyRes);
                    //foreach (var testElem in list.List) 
                    //{
                    //    dist = FuzzyString.ComparisonMetrics.HammingDistance(temp, testElem);
                    //    if (dist == 0)
                    //    {
                    //        fuzzyRes.BestMatch = testElem;
                    //        fuzzyRes.Classification = list.Classification;
                    //        returningList.Add(fuzzyRes);
                    //        found = true;
                    //        break;
                    //    }
                    //    if (dist <= 2)
                    //    {
                    //    }


                    //}
                    //if (found)
                    //    break;
                }

                i++;
                //if (found)
                //    continue;
                #region Last names and first names, needs to be split before fuzzy searching
                //foreach (var list in gazetteerLists.Where(g => g.Classification == "Last name" || g.Classification == "First name"))
                //{
                //    temp = elem;
                //    if (list.Classification == "Last name" && elem.Contains(' ') && elem.Split(' ').Last() != "")
                //        temp = elem.Split(' ').Last();
                //    if (list.Classification == "First name" && elem.Contains(' ') && elem.Split(' ').First() != "")
                //        temp = elem.Split(' ').First();
                //    foreach (var testElem in list.List)
                //    {
                //        dist = FuzzyString.ComparisonMetrics.HammingDistance(temp, testElem);
                //        if (dist == 0)
                //        {
                //            fuzzyRes.OriginalString = temp;
                //            fuzzyRes.BestMatch = testElem;
                //            fuzzyRes.Classification = list.Classification;
                //            returningList.Add(fuzzyRes);
                //            found = true;
                //            break;
                //        }
                //        if (dist <= 2)
                //        {
                //        }


                //    }
                //}
                #endregion
            }

            return returningList;

        }


        //Testing with firstname/lastname
        public List<LevenshteinResult> CustomLevenshteinDistanceTest(IEnumerable<string> testingList, IEnumerable<string> oilCompPosTitles, IEnumerable<string> allLastnames, IEnumerable<string> allFirstnames)
        {
            var testingListCount = testingList.Count();
            bool bApproxFound = false;
            int approxFound = 0;
            int found = 0;
            var split = "";
            var levDist = 0;
            var i = 0;
            List<LevenshteinResult> bestMatches = new List<LevenshteinResult>();
            foreach (var row in testingList)
            {
                LevenshteinResult temp = new LevenshteinResult(originalString: row);
                Console.Title = (int)(((double)i / (double)testingListCount) * 100) + "% - MasterConsole";
                Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                bApproxFound = false;
                Console.WriteLine("Checking for: " + row);
                split = row;
                foreach (var posTitle in oilCompPosTitles)
                {
                    levDist = LevenshteinDistance.Compute(posTitle, row);
                    if (levDist == 0)
                    {
                        bApproxFound = false;
                        found++;
                        Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                        Console.WriteLine("Position title match found for '" + row + "': " + posTitle);
                        temp.BestMatch = posTitle;
                        temp.LevenshteinDistance = levDist;
                        temp.Classification = Classification.PositionTitle;
                        break;
                    }

                    if (levDist < 2)
                    {
                        bApproxFound = true;
                        Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                        Console.WriteLine("Position title approx match found for '" + row + "': " + posTitle);
                        if (levDist < temp.LevenshteinDistance)
                        {
                            temp.BestMatch = posTitle;
                            temp.LevenshteinDistance = levDist;
                            temp.Classification = Classification.PositionTitle;
                        }
                    }
                }
                if (temp.LevenshteinDistance > 0)
                {
                    if (row.Contains(' ') && row.Split(' ').Last() != "")
                        split = row.Split(' ').Last();
                    foreach (var lastname in allLastnames)
                    {
                        levDist = LevenshteinDistance.Compute(lastname, split);
                        if (levDist == 0)
                        {
                            bApproxFound = false;
                            found++;
                            Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                            Console.WriteLine("Last name match found for '" + row + "': " + lastname);
                            temp.BestMatch = row.Replace(split, lastname);
                            temp.LevenshteinDistance = levDist;
                            temp.Classification = Classification.Name;
                            break;
                        }

                        if (levDist < 2)
                        {
                            bApproxFound = true;
                            Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                            Console.WriteLine("Last name approx match found for '" + row + "': " + lastname);
                            if (levDist < temp.LevenshteinDistance)
                            {
                                temp.BestMatch = row.Replace(split, lastname);
                                temp.LevenshteinDistance = levDist;
                                temp.Classification = Classification.Name;
                            }
                        }
                    }
                    split = row;
                    if (row.Contains(' ') && row.Split(' ').First() != "")
                        split = row.Split(' ').First();
                    if (temp.BestMatch != string.Empty && temp.BestMatch.Contains(' ') && temp.Classification == Classification.Name)
                        split = temp.BestMatch.Split(' ').First();
                    foreach (var firstname in allFirstnames)
                    {
                        levDist = LevenshteinDistance.Compute(firstname, split);
                        if (levDist == 0)
                        {
                            bApproxFound = false;
                            found++;
                            Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                            Console.WriteLine("First name match found for '" + row + "': " + firstname);
                            temp.BestMatch = row.Replace(split, firstname);
                            temp.LevenshteinDistance = levDist;
                            temp.Classification = Classification.Name;
                            break;
                        }

                        if (levDist < 2)
                        {
                            bApproxFound = true;
                            Console.Write("\r{0}%    ", (int)(((double)i / (double)testingListCount) * 100));
                            Console.WriteLine("First name approx match found for '" + row + "': " + firstname);
                            if (levDist < temp.LevenshteinDistance)
                            {
                                temp.BestMatch = row.Replace(split, firstname);
                                temp.LevenshteinDistance = levDist;
                                temp.Classification = Classification.Name;
                            }
                        }
                    }
                }
                if (bApproxFound)
                    approxFound++;
                bestMatches.Add(temp);
                i++;
            }
            return bestMatches;
        }


        /// <summary>
        /// A string-list of the elements in the column.
        /// Resources used to create this method: http://stackoverflow.com/questions/13886330/read-excel-first-column-using-c-sharp-into-array
        /// </summary>
        /// <param name="worksheetNumber">The worksheet number (starts at 1)</param>
        /// <param name="columnNumber">The column number (starts at 1)</param>
        /// <param name="filename">The path and name of the file</param>
        /// <returns>A string-list of the elements in the column</returns>
        public static IEnumerable<string> GetColumnList(int worksheetNumber, int columnNumber, string filename)
        {
            Microsoft.Office.Interop.Excel.Application xlsApp = new Microsoft.Office.Interop.Excel.Application();
            if (xlsApp == null)
            {
                Console.WriteLine("EXCEL could not be started. Check that your office installation and project references are correct.");
                return null;
            }
            Microsoft.Office.Interop.Excel.Workbook wb = xlsApp.Workbooks.Open(filename,
            0, true, 5, "", "", true, XlPlatform.xlWindows, "\t", false, false, 0, true);
            Microsoft.Office.Interop.Excel.Sheets sheets = wb.Worksheets;
            Microsoft.Office.Interop.Excel.Worksheet ws = (Microsoft.Office.Interop.Excel.Worksheet)sheets.get_Item(worksheetNumber);

            Range column = ws.UsedRange.Columns[columnNumber];
            System.Array myvalues = (System.Array)column.Cells.Value;
            var strArray = myvalues.OfType<object>().Select(o => o.ToString());
            wb.Close(0);
            xlsApp.Quit();
            System.Runtime.InteropServices.Marshal.ReleaseComObject(xlsApp);
            return strArray;
        }

        //public static void WriteToExcel(List<LevenshteinObject> list, string filename)
        //{
        //    Microsoft.Office.Interop.Excel.Application xlsApp = new Microsoft.Office.Interop.Excel.Application();
        //    xlsApp.Visible = true;
        //    xlsApp.Workbooks.Add();
        //    if (xlsApp == null)
        //    {
        //        Console.WriteLine("EXCEL could not be started. Check that your office installation and project references are correct.");
        //    }
        //    Microsoft.Office.Interop.Excel.Workbook wb = xlsApp.Workbooks.Open(filename,
        //    0, false, 5, "", "", false, XlPlatform.xlWindows, "", true, false, 0, true, false, false);
        //    Microsoft.Office.Interop.Excel.Sheets sheets = wb.Worksheets;
        //    Microsoft.Office.Interop.Excel.Worksheet ws = (Microsoft.Office.Interop.Excel.Worksheet)sheets.get_Item(worksheetNumber);

        //    Range column = ws.UsedRange.Columns[columnNumber];
        //    System.Array myvalues = (System.Array)column.Cells.Value;
        //    var strArray = myvalues.OfType<object>().Select(o => o.ToString());
        //    wb.Close(0);
        //    xlsApp.Quit();
        //    System.Runtime.InteropServices.Marshal.ReleaseComObject(xlsApp);
        //    return strArray;
        //}

        // The DOM approach.
        // Note that the code below works only for cells that contain numeric values.
        // 
        public static void ReadExcelFileDOM(string fileName)
        {
            using (SpreadsheetDocument spreadsheetDocument = SpreadsheetDocument.Open(fileName, false))
            {
                WorkbookPart workbookPart = spreadsheetDocument.WorkbookPart;
                WorksheetPart worksheetPart = workbookPart.WorksheetParts.First();
                SheetData sheetData = worksheetPart.Worksheet.Elements<SheetData>().First();
                string text;
                foreach (Row r in sheetData.Elements<Row>())
                {
                    foreach (Cell c in r.Elements<Cell>())
                    {
                        text = c.CellValue.Text;
                        Console.Write(text + " ");
                    }
                }
                Console.WriteLine();
                Console.ReadKey();
            }
        }

        // The SAX approach.
        public static List<string> ReadExcelFileSAX(string fileName)
        {
            using (SpreadsheetDocument spreadsheetDocument = SpreadsheetDocument.Open(fileName, false))
            {
                var list = new List<string>();
                WorkbookPart workbookPart = spreadsheetDocument.WorkbookPart;
                //WorksheetPart worksheetPart = workbookPart.WorksheetParts.First();
                workbookPart.Workbook.Sheets.ToList().ForEach(x => Console.WriteLine(x.GetAttribute("name", "").Value));
                
                var worksheetId = workbookPart.Workbook.Descendants<Sheet>().First(s => s.Name == "Last Names").Id;

                WorksheetPart worksheetPart = (WorksheetPart)workbookPart.GetPartById(worksheetId);

                OpenXmlReader reader = OpenXmlReader.Create(worksheetPart);
                string text;
                while (reader.Read())
                {
                    if (reader.ElementType == typeof(CellValue))
                    {
                        text = reader.GetText();
                        list.Add(text);
                        //Console.Write(text + " ");
                    }
                }
                return list;
                //Console.WriteLine();
                //Console.ReadKey();
            }
        }



    }
}
